/*-------------------------------------------------------------------------
# Author: Federico Guerra
# Date: 4 July 2007
# Department of Information Engineering (DEI), University of Padova
#-------------------------------------------------------------------------
#
# Copyright (c) 2007 Regents of the SIGNET lab, University of Padova.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of the University of Padova (SIGNET lab) nor the 
#    names of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED 
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*---------------------------------------------------------------------------
* Description: 
* This module simulates a voice monologue implementing the Gruber Model
*---------------------------------------------------------------------------
* Example of usage in TCL:
* (for the complete list of "set" check cnv_defaults.tcl)* 
*
* set mnlg_example    [new Module/MNLG]
*
* $mnlg_example set framerate_ 0.020
* $mnlg_example set hangover_ 0
* $mnlg_example set fillin_ 0
* $mnlg_example set seed_ 222
*--------------------------------------------------------------------------- 
* TCL Command list (for a command explanation search this file)
*
* $mnlg_example start 
* $mnlg_example start_with_VAD
* $mnlg_example start_with_VAD_hangover
* $mnlg_example start_with_VAD_fillin
* $mnlg_example start_with_fillin
* $mnlg_example start_with_hangover
*-------------------------------------------------------------------------*/

#include <iostream>
#include <cmath>
#include <node-core.h>
#include <ip.h>

#include "mnlg-module.h"


extern packet_t PT_MNLG;

int hdr_monologue::offset_;

static class MonologuePktClass : public PacketHeaderClass {
public:
  MonologuePktClass() : PacketHeaderClass("PacketHeader/MNLG", sizeof(hdr_monologue)) {
    this->bind();
    bind_offset(&hdr_monologue::offset_);
  }
} class_monologue_pkt; 



static class MonologueModuleClass : public TclClass {
public:
  MonologueModuleClass() : TclClass("Module/MNLG") {}
  TclObject* create(int, const char*const*) {
    return (new MonologueModule());
  }
} class_module_monologue;


void SendTimer::expire(Event *e)
{
  module_send->transmit();
}

void StateTimer::expire(Event *e)
{ // we are checking if number of pcks to send is an integer
  double frame_remainder = fmod(module_state->statetime_ , module_state->framerate_);
  if (frame_remainder == 0.0) // if pcks is a perfect integer then change state
  {
     module_state->changeState();
  }  
  else { // else we have to send the remaining packet
     module_state->transmit(); 
     module_state->changeState();
  }
}

int MonologueModule::uidcnt_ ;		// unique id of the packet generated      

/*Gruber parameters initialization*/

double MonologueModule::lambda1 = 6.322;
double MonologueModule::lambda2 = 27.62;
double MonologueModule::lambda3 = 1.377;
double MonologueModule::R2 = 0.8764 ;
double MonologueModule::R3 = 0.1469 ;


MonologueModule::MonologueModule()  
  : sendTmr_(this),
    stateTmr_(this),
    statetime_(0),
    txsn(1),
    hrsn(0),
    state_(0),
    pkts_recv(0),
    pkts_ooseq(0),
    pkts_lost(0),
    pkts_last_reset(0),
    rftt(-1),
    lrtime(0),
    sthr(0)
{
  // binding to TCL variables
  bind("framerate_", &framerate_);
  bind("destPort_", &dstPort_);
  bind("destAddr_", &dstAddr_);
  bind("packetSize_", &pktSize_);
  bind("alpha_", &alpha_);
  bind("hangover_", &hangover_);
  bind("fillin_", &fillin_);
  bind("seed_", &seed_);
  uidcnt_ = 0;
}

MonologueModule::~MonologueModule()
{
}

// TCL command interpreter
int MonologueModule::command(int argc, const char*const* argv)
{
  Tcl& tcl = Tcl::instance();
  
  if(argc==2)
    {
      if(strcasecmp(argv[1], "start") == 0)	/* TCL command to start the packet generation and transmission
	                                         * without VAD,nor hangover nor fillin ==> CBR */
	{
	  hangover_ON = false;
          vad_ON = false;
	  fillin_ON = false;
	  start();
	  return TCL_OK;
	}
      else if(strcasecmp(argv[1], "stop") == 0)	// TCL command to stop the packet generation
	{
	  stop();
	  return TCL_OK;
	}
      else if(strcasecmp(argv[1], "getrtt") == 0)	
	{
	  tcl.resultf("%f",GetRTT());
	  return TCL_OK;
	}
      else if(strcasecmp(argv[1], "getftt") == 0)	
	{
	  tcl.resultf("%f",GetFTT());
	  return TCL_OK;
	}
      else if(strcasecmp(argv[1], "getper") == 0)	
	{
	  tcl.resultf("%f",GetPER());
	  return TCL_OK;
	}	
      else if(strcasecmp(argv[1], "getthr") == 0)	
	{
	  tcl.resultf("%f",GetTHR());
	  return TCL_OK;
	}
      else if(strcasecmp(argv[1], "sendPkt") == 0)	
	{
	  sendPkt();
	  return TCL_OK;
	}
      else if(strcasecmp(argv[1], "nextState") == 0)	
	{
	  nextState();
	  return TCL_OK;
	}
      else if(strcasecmp(argv[1], "start_with_hangover") == 0) // command to start with hangover
	{
	  vad_ON = false;
          hangover_ON = true;
	  fillin_ON = false;
	  start();
        return TCL_OK;
	}
     else if(strcasecmp(argv[1], "start_with_VAD") == 0) //command to start with VAD
	{
	  hangover_ON = false;
          vad_ON = true;
          fillin_ON = false;
          start();
	  return TCL_OK;
	}
     else if(strcasecmp(argv[1], "start_with_VAD_hangover") == 0) // command to start with VAD,hangover
	{
	  vad_ON = true;
          hangover_ON = true;
	  fillin_ON = false;
          start();
	  return TCL_OK;
	}
     else if(strcasecmp(argv[1], "start_with_VAD_fillin") == 0)	//command to start with VAD,fillin
	{
	  vad_ON = true;
          hangover_ON = false;
	  fillin_ON = true;
          start();
	  return TCL_OK;
	}
     else if(strcasecmp(argv[1], "start_with_fillin") == 0)	// command to start with fillin
	{
	  vad_ON = false;
          hangover_ON = false;
	  fillin_ON = true;
          start();
	  return TCL_OK;
	}
     else if(strcasecmp(argv[1], "resetStats") == 0)	
	{
	  resetStats();
	  fprintf(stderr,"MonologueModule::command() resetStats %s, pkts_last_reset=%d, hrsn=%d, txsn=%d\n", tag_ , pkts_last_reset, hrsn, txsn);
	  return TCL_OK;
	}
    }

  return Module::command(argc, argv);
}


int MonologueModule::crLayCommand(ClMessage* m)
{
  switch (m->type()) 
    {
    
      //case whatever: 
      //    return 0;
      //    break;
    
    default:
      return Module::crLayCommand(m);    
    }  
}

int MonologueModule::getState() 
{
  return state_;
}


void MonologueModule::initPkt(Packet* p)
{
  hdr_cmn* ch = hdr_cmn::access(p);
  ch->uid() = uidcnt_++;
  ch->ptype() = PT_MNLG;
  ch->size() = pktSize_;
  
  hdr_ip* iph = hdr_ip::access(p);
  iph->daddr() = dstAddr_;
  iph->dport() = dstPort_;

  hdr_monologue* mnlgh = HDR_MNLG(p);
  mnlgh->sn = txsn;
  txsn++;
  mnlgh->ts = Scheduler::instance().clock();
  ch->timestamp() = mnlgh->ts;
  
  // how to set the is_voice flag
  
  if (state_ == 0 ) // we are in the "voice" state
    {
    mnlgh->is_voice = true; // the packet is tagged "voice"
    hangover_counter = hangover_pcks; /*reset the hangover counter for the
                                       *next transition to state 1*/
	}
  else if ( (state_ == 1) && (hangover_ON == true) && (hangover_counter > 0) && (fillin_ON == false) )
   { // we are in "silence" state BUT hangover is ON and it still has pcks to attach
    mnlgh->is_voice = true;
    hangover_counter--; // decrease the counter for the next call of inintPck()
   }
  else if ( (state_ == 1) && (hangover_ON == true) && (hangover_counter <= 0) && (fillin_ON == false) )
   { // hangover is ON but there are no pcks left to attach
    mnlgh->is_voice = false;
   } 
  else if ( (state_ == 1) && (hangover_ON == false) && (fillin_ON == false) ) 
  { // we are in "silence" state with no hangover nor fillin
   mnlgh->is_voice = false;
  }
  else if ( (state_ == 1) && (hangover_ON == false) && (fillin_ON == true) )
  { //we are in "silence" state but fillin is on
   
   /*hack so we can know in advance how many "silence" pcks will be generated*/
   int silence_pcks = (int)ceil(statetime_/framerate_);
   int fillin_pcks = (int)ceil(fillin_/framerate_); //number of max fillin pcks
   
   if (fillin_pcks >= silence_pcks) 
   { // the fillin can cover the "silence" period
    mnlgh->is_voice = true;
   }
   else
   { // the fillin can cover the "silence" period
    mnlgh->is_voice = false;
   }
  }
  else 
  {
  printf("MonologueModule(%d)::initPck() Error in is_voice initialization \n",getId());
  exit(1);
  }
  if (rftt >= 0)
    {
      mnlgh->rftt = rftt;
      mnlgh->rftt_valid = true;
    }
  else
    {
      mnlgh->rftt_valid = false;
    }

}

void MonologueModule::setStateTime() 
{
   double y = generator1.next_double(); // next uniform double bewteen (0,1)
   statetime_ = ((-1)*log(y)/lambda1) ; //generates the markov chain's state waiting time
   if (debug_ > 0) {
   printf("MonologueModule(%d)::setStateTime() statetime_ of (%d) = (%f)\n",getId(),state_,statetime_);	
   }
}

void MonologueModule::nextState()
{
  if (state_ == 1) {
	
        state_ = 0; // we only have two states so we don't need any state transition probability
        double y = generator1.next_double();
	statetime_ = ((-1)*log(y)/lambda1) ; 
	if (statetime_ < framerate_) //We can't allow statetime_ < time resolution
        {                            
            statetime_ = framerate_; // so we force to last at least a single frame
        }	
	if (debug_ > 0) {
        printf("MonologueModule(%d)::nextState() statetime_ of (%d) = (%f)\n",getId(),state_,statetime_);
        }
  }
  else if (state_ == 0)  {
        
        state_ = 1;
	double x = generator1.next_double();	
	/*we split the decision into two: first we decide if we are in short or long pause, then we calculate
	 *the waiting time */	
	if ( x <= R2 ) { //case 1
		double y = generator1.next_double();
		statetime_ = min(1.4 , ((-1)*log(y)/lambda2) ) ; // the min is imposed by gruber model
		if (statetime_ < framerate_) //We can't allow statetime_ < time resolution
                {                            
                statetime_ = framerate_;
                }
		if (debug_ > 0) {
        	printf("MonologueModule(%d)::nextState() statetime_ of (%d) = (%f)\n",getId(),state_,statetime_);
        	}
	}
	else if ( x > R2 ) { //case 2
		double y = generator1.next_double();
		statetime_ = min(1.4 , ((-1)*log(y)/lambda3) ) ;
		if (statetime_ < framerate_) //We can't allow statetime_ < time resolution
                {                            
                statetime_ = framerate_;
                }		
                if (debug_ > 0) {
        	printf("MonologueModule(%d)::nextState() statetime_ of (%d) = (%f)\n",getId(),state_,statetime_);
        	}	
	}
	else {
		if (debug_ > 0) {
        	printf("MonologueModule(%d)::nextState() statetime_ of (%d) = (%f)\n",getId(),state_,statetime_);
        	}
             }
	}
	else {
		printf("MonologueModule(%d)::nextState() unknow error in state transition \n",getId());
		exit(1);		
	}
}

void MonologueModule::start()
{
  /* we check the given parameters */
  if ((framerate_ <= 0) || (hangover_ < 0) || (fillin_ < 0) )
    {
      fprintf(stderr,"MonologueModule(%d)::start() a parameter is <= 0 \n",getId());
      exit(1);
    }
  /* generators intialization*/
    //  generator1.set_seed(seed_);
  for (int i = 1; i < seed_; i++) {
    generator1.reset_next_substream();
    //generator2.reset_next_substream();
  }
  /* hangover initialization */
  hangover_pcks = (int)ceil(hangover_/framerate_); 
  hangover_counter = 0;
  setStateTime(); // we calculate the first waiting time (statetime_)
  sendTmr_.resched(framerate_); //we reschedule the pck timer
  
  /* if VAD is on we also turn on the markov chain: this is intended to preserve the
   * the cbr behaviour if the VAD is off */
  if (vad_ON == true) stateTmr_.resched(statetime_); 
}

void MonologueModule::sendPkt()
{
  double delay;
  
  if (fillin_ON == false) { 
     delay = 0;
  }
  else {
     delay = fillin_;
  }
  Packet* p = Packet::alloc();
  initPkt(p);
  hdr_cmn* ch = hdr_cmn::access(p);
  hdr_monologue* mnlgh = HDR_MNLG(p);
  if (debug_>0)
    printf("MonologueModule(%d)::sendPkt(), send a pkt (%d)\n",getId(), ch->uid());
  
  /* VAD SECTION */
  if ( (vad_ON == true) && (mnlgh->is_voice == true) ) // pck is voice => we can send it
  {
  sendDown(p,delay);
  }
  else if ( (vad_ON == true) && (mnlgh->is_voice == false))// pck is silence => we discard it
  {
  if (debug_>0)
    {
    printf("MonologueModule(%d)::sendPkt(), dropped pkt by VAD \n",getId(), ch->uid());
    }
  Packet::free(p);
  }
  else if ( vad_ON == false ) // if the VAD is off we send the pck down
  {
  sendDown(p,delay);
  }
  else 
  {
  printf("MonologueModule(%d)::sendPkt(), error occured in VAD section \n",getId());  
  exit(1);
  }

}

void MonologueModule::changeState() 
{
  sendTmr_.cancel(); // we stop the current pck timer
  nextState(); // we do a a state transition
  sendTmr_.resched(framerate_); // we reschedule the pck timer
  stateTmr_.resched(statetime_); // and then reschedule the state time with the new waiting state time
  
}

void MonologueModule::transmit()
{
  if (debug_ > 0) {
     printf("MonologueModule(%d)::transmit() state %d , hangover counter %d \n",getId(),state_ ,hangover_counter);
  }
  sendPkt();
  /* schedule next trasnmission */
  sendTmr_.resched(framerate_);
}

void MonologueModule::stop()
{
  sendTmr_.cancel();
  
  /*if VAD is ON the we stop the state timer*/
  if (vad_ON == true) stateTmr_.cancel();
}


void MonologueModule::recv(Packet* p, Handler* h)
{
  hdr_cmn* ch = hdr_cmn::access(p);
  recv(p);
}

/* same method of Miracle CbrModule.cc */
void MonologueModule::recv(Packet* p)
{
  hdr_cmn* ch = hdr_cmn::access(p);
  if (debug_>0)     
    printf("MonologueModule(%d)::recv(Packet*p,Handler*) pktId %d\n",getId(), ch->uid());
    
  if (ch->ptype() != PT_MNLG)
    {
      drop(p,1,MNLG_DROP_REASON_UNKNOWN_TYPE);
      incrPktInvalid();
      return;
    }

  hdr_ip* iph = hdr_ip::access(p);
  if (iph->saddr() != dstAddr_)
    {
      drop(p,1,MNLG_DROP_REASON_WRONG_SADDR);
      incrPktInvalid();
      return;
    }
  if (iph->sport() != dstPort_)
    {
      drop(p,1,MNLG_DROP_REASON_WRONG_SPORT);
      incrPktInvalid();
      return;
    }

  hdr_monologue* mnlgh = HDR_MNLG(p);

  int esn = hrsn + 1; // expected sn
    
  if (mnlgh->sn < esn)
    { // packet is out of sequence and is to be discarded
      incrPktOoseq();
      if (debug_ > 0)
	printf("MnlgModule::recv() Pkt out of sequence! mnlgh->sn=%d\thrsn=%d\tesn=%d\n",mnlgh->sn,hrsn,esn);
      drop(p,1,MNLG_DROP_REASON_OUT_OF_SEQUENCE);
      return;
    }
  
  /* a new packet has been received */
  incrPktRecv();
  hrsn = mnlgh->sn;  

  if (mnlgh->sn > esn)
    { // packet losses are observed
      incrPktLost(mnlgh->sn - (esn));     
    }
  
  /* Smoothed throughput calculation */
  double dt = std::max((Scheduler::instance().clock() - lrtime), 1e-6);
  double thr = ch->size() / dt;
  updateThroughput(thr);
  lrtime = Scheduler::instance().clock(); 
    
  /* Forward and Round Trip Time calculation */
  rftt = Scheduler::instance().clock() - mnlgh->ts;
  if(mnlgh->rftt_valid) 
    {
      updateFTT(mnlgh->rftt);
      double rtt = rftt + mnlgh->rftt;
      updateRTT(rtt);
    }

  //drop(p,5,MNLG_DROP_REASON_RECEIVED);   
  Packet::free(p);

  if (pkts_lost + pkts_recv + pkts_last_reset != hrsn)     
    fprintf(stderr,"ERROR MonologueModule::recv() pkts_lost=%d  pkts_recv=%d  hrsn=%d\n", pkts_lost, pkts_recv, hrsn);

  //assert(pkts_lost + pkts_recv == hrsn);
  
}

/* same methods of Miracle CbrModule.cc */

double MonologueModule::GetRTT() 
{
  return srtt;
}


double MonologueModule::GetFTT() 
{
  return sftt;
}



double MonologueModule::GetPER() 
{
  if ((pkts_recv + pkts_lost)> 0)
    return ((double) pkts_lost / (double)(pkts_recv + pkts_lost));
  else 
    return 0;
}


double MonologueModule::GetTHR() 
{
  return sthr;
}


void MonologueModule::updateRTT(double rtt)
{
      srtt = alpha_ * srtt + (1-alpha_) * rtt;
}

void MonologueModule::updateFTT(double ftt)
{
  sftt = alpha_ * sftt + (1-alpha_) * ftt;
}

void MonologueModule::updateThroughput(double thr)
{
  sthr = alpha_ * sthr + (1-alpha_) * thr;
}

void MonologueModule::incrPktLost(int npkts)
{
  pkts_lost += npkts;
}

void MonologueModule::incrPktRecv()
{
  pkts_recv++;
}

void MonologueModule::incrPktOoseq()
{
  pkts_ooseq++;
}

void MonologueModule::incrPktInvalid()
{
  pkts_invalid++;
}


void MonologueModule::resetStats()
{
  pkts_last_reset += pkts_lost + pkts_recv;
  pkts_recv = 0;
  pkts_ooseq = 0;
  pkts_lost = 0;
  srtt = 0;
  sftt = 0;
  sthr = 0;
  //txsn = 1;
  //hrsn=0;
  rftt = -1;
}

