/*-------------------------------------------------------------------------
# Author: Federico Guerra
# Date: 4 July 2007
# Department of Information Engineering (DEI), University of Padova
#-------------------------------------------------------------------------
#
# Copyright (c) 2007 Regents of the SIGNET lab, University of Padova.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of the University of Padova (SIGNET lab) nor the 
#    names of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED 
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.*/

#ifndef MONOLOGUE_MODULE
#define MONOLOGUE_MODULE

#include <module.h>
#include <rng.h>

/* same definitions of Miracle CbrModule.h */
#define MNLG_DROP_REASON_UNKNOWN_TYPE "UKT"
#define MNLG_DROP_REASON_RECEIVED "RCV"
#define MNLG_DROP_REASON_OUT_OF_SEQUENCE "OOS"
#define MNLG_DROP_REASON_WRONG_SADDR "WSA"
#define MNLG_DROP_REASON_WRONG_SPORT "WSP"

#define HDR_MNLG(P)      (hdr_monologue::access(P))

extern packet_t PT_MNLG;

typedef struct hdr_monologue {

  double ts;  /**< packet timestamp, i.e., its generation time) */
  int sn;     /**< sequence number of this packet */  
  bool rftt_valid; /**< whether the rftt value is valid (a packet must
		    * have been received in the opposite direction) 
		    */ 
  bool is_voice;  // to tag if its type is voice or silence

  double rftt;    /**< Forward Trip Time seen in the opposite direction
		* (opposite with respect to the direction of this
		* packet). This is used to calculate Round Trip Time 
		*/
  
/*same instructions of Miracle CbrModule.h */
  static int offset_;
  inline static int& offset() { return offset_; }
  inline static struct hdr_monologue* access(const Packet* p) {
    return (struct hdr_monologue*)p->access(offset_);
  }
  
} hdr_monologue;


class MonologueModule;


/**
 * Timer used by MnlgModule to schedule packet transmission
 */
class SendTimer : public TimerHandler
{
 public:
  SendTimer(MonologueModule *m) : TimerHandler() { module_send = m; }
	
 protected:
  virtual void expire(Event *e);
  MonologueModule* module_send;
};

/**
 * Timer used by MnlgModule to schedule Semi-Markov Process's state transition  
 */
class StateTimer : public TimerHandler
{
 public:
  StateTimer(MonologueModule *m) : TimerHandler() { module_state = m; }
	
 protected:
  virtual void expire(Event *e);
  MonologueModule* module_state;
};


/**
 * Implements a Monologue sender/receiver model based on Gruber paper 
 */
class MonologueModule : public Module
{
  friend class SendTimer;
  friend class StateTimer;
  
public:

  MonologueModule();
  ~MonologueModule();		
		
  virtual void recv(Packet*);
  virtual void recv(Packet* p, Handler* h);
  virtual int command(int argc, const char*const* argv);
  virtual int crLayCommand(ClMessage* m);

  /*same methods of Miracle CbrModule.h */
  virtual double GetRTT();
  virtual double GetFTT();
  virtual double GetPER();  
  virtual double GetTHR(); 

  virtual void resetStats(); 


protected:

  int seed_; // seed for the internal RNG generators
  double statetime_; /** Semi-Markov Process's waiting time until next transition */
  int state_;   /**  Semi-Markov Process's state ={0,1}  */
  static int uidcnt_;  /**< simulation-unique packet ID */
  nsaddr_t dstPort_;
  nsaddr_t dstAddr_;
  
  double hangover_; // hangover time in secs
  double fillin_; //fillin time in secs
  bool vad_ON;      //flag for internal case switching
  bool hangover_ON; //flag for internal case switching
  bool fillin_ON; //flag for internal case switching
  
  int hangover_pcks; // number of total hangover pcks
  int hangover_counter; //counter for processed hangover pcks
  
  RNG generator1; // random generator for the Semi-Markov Process
  //RNG generator2; // random generator for the Semi-Markov Process
  
  /**
   * Generate a packet and transmit it
   */ 
  virtual void sendPkt();
  
  /**
   * Calls sendPkt and schedule a new transmission after period_ seconds
   */
  virtual void transmit();
  
  /** 
  * calls nextState and schedules a new transition after
  * a statetime_ seconds
  */
  virtual void changeState();
  
  /** 
  * does a transition in the Semi-Markov Process
  */
  virtual void nextState();
  
  /*
  * calculates the waiting time for the current state_
  */
  virtual void setStateTime();
  
  /*
  * returns current state_
  */
  virtual int getState();

  
  /** 
   * initialize a packet before transmission
   * 
   * @param p packet to be initialized
   */	
  virtual void initPkt(Packet* p);

  /** 
   * start sending packets 
   */
  virtual void start();
		

  /** 
   * stop sending packets 
   */
  virtual void stop();

  /*same methods of Miracle CbrModule.h*/
  virtual void updateRTT(double rtt);
  virtual void updateFTT(double ftt);
  virtual void updateThroughput(double thr);

  virtual void incrPktLost(int npkts);
  virtual void incrPktRecv();
  virtual void incrPktOoseq();
  virtual void incrPktInvalid();

  double alpha_;      /**< smoothing factor used for statistics using
                          first order IIR filter */
	       
  static double lambda1; /* parameter of the gruber model for the calculation of the waiting time
                          *  for the calculation of the waiting time */
  static double lambda2; /* parameter of the gruber model for the calculation of the waiting time
                          *  for the calculation of the waiting time */
  static double lambda3; /* parameter of the gruber model for the calculation of the waiting time
                          *  for the calculation of the waiting time */
  static double R2; /* parameter of the gruber model for the calculation of the waiting time
                          *  for the calculation of the waiting time */
  static double R3; /* parameter of the gruber model for the calculation of the waiting time
                          *  for the calculation of the waiting time */
  
  int txsn;          /**< sequence number of next new packet to be transmitted */
  int hrsn;          /**< highest received sequence number */

  int pkts_recv;     /**< total number of received packets. Packet out
			of sequence are not counted here. */
  int pkts_ooseq;    /**< total number of packets received out of
			sequence  */
  int pkts_lost;     /**< total number of lost packets, including
			packets received out of sequence   */

  int pkts_invalid;  /**< total number of invalid packets received */

  int pkts_last_reset; /**< for error checking after stats are
			  reset. Set to pkts_lost+pkts_recv each time
			  resetStats is called. */ 

  double rftt;       /**< Forward Trip Time seen for last received
			packet */ 

  double srtt;       /**< Smoothed Round Trip Time, calculated as for
			TCP */

  double sftt;       /**< Smoothed Forward Trip Time, calculated as srtt */

  double lrtime;       /**< Time of last packet reception */
  double sthr;       /**< Smoothed throughput calculation */

  double framerate_;      /**< period between two consecutive packet transmissions */
  
  int pktSize_;        /**< packet size */
  
  SendTimer sendTmr_;  /**< timer which schedules packet transmissions */
  
  StateTimer stateTmr_ ; /** < timer which schedules model state-transitions */
		
};




#endif /* MONOLOGUE_MODULE */
 
